#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `pynrc` package."""

import pytest
#import pynrc


def test_the_obvious():
    print('pyNRC testing not yet implemented!!')
    assert True == True


if __name__ == '__main__':
    pytest.main()
